import importlib

importlib.import_module("neurst.data.audio")
importlib.import_module("neurst.data.data_pipelines")
importlib.import_module("neurst.data.datasets")
importlib.import_module("neurst.data.text")
